--[[
_ Talented is a complete rewrite of the Talent Frame using a little bit of Ace.

_ Talented makes a lot more assumptions than the original Blizzard code about talents.

_ Talented show all 3 talent tabs at the same time.
_ Talented allows to create and store templates
]]

Talented = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceDebug-2.0", "AceHook-2.1", "AceEvent-2.0")

Talented:RegisterDB("TalentedDB", "TalentedCharDB")
Talented:RegisterDefaults("class", { templates = {} })

function Talented:InitializeOptions()
	self.options = {
		desc = "Talented - Talent Editor",
		type = "group",
		args = {
			mode = {
				name = "Mode",
				desc = "Mode of operation",
				type = "text",
				usage = "<mode>",
				get = function() return self.mode end,
				set = function(value) self:SetMode(value) end,
				validate = { "view", "edit", "apply", "delete" },
			},
			template = {
				name = "Template",
				desc = "Template",
				type = "group",
				args = {
					new = {
						name = "New Template",
						desc = "Create a new Template",
						type = "group",
						order = 100,
						args = {
							empty = {
								name = "New empty template",
								desc = "Create a new template from scratch",
								type = "execute",
								order = 100,
								func = function() self:SetTemplate(self:CreateEmptyTemplate()) end,
							},
							h1 = {
								type = "header",
								order = 150,
							},
							-- copy from existing templates here
							current = {
								name = "Copy from current",
								desc = "Create a new template from the current",
								type = "execute",
								order = 300,
								func = function() self:SetTemplate(self:CopyTemplate(self.current)) end,
							},
						},
					},
					h1 = {
						type = "header",
						order = 150,
					},
					-- copy from existing templates here
					current = {
						name = "Current template",
						desc = "Select the current template",
						type = "execute",
						order = 300,
						func = function() self:SetTemplate(self.current) end,
					},
				}
			}
		},
	}
	
	do
		local src = self.db.char.templates
		local dst = self.db.class.templates
		if src then
			for name, template in pairs(src) do
				if not dst[name] then
					dst[name] = template
					src[name] = nil
				end
			end
		end
	end
	
	for _, template in pairs(self.db.class.templates) do
		self:AddTemplateToOptions(template)
	end
end

function Talented:GetMode()
	return self.mode
end

function Talented:ApplyCurrentTemplate()
	local template = self.template
	local count = 0
	-- check if enough talent points are available
	local available = UnitCharacterPoints("player")
	for tab, talents in ipairs(self.talents) do
		for index, info in ipairs(talents.infos) do
			local delta = template[tab][index] - self.current[tab][index]
			if delta > 0 then
				count = count + delta
			end
		end
	end
	if count > available then
		self:Print("Sorry, I can't apply this template because you don't have enough talent points available (need %d)", count)
		self.mode = "view"
		self:UpdateView()
	else
		-- let's go !
		self:Print("Please wait while I set your talents !")
		self:EnableUI(false)
		self:CHARACTER_POINTS_CHANGED()
	end
end

function Talented:SetMode(mode)
	if self.template == self.current then
		mode = "view"
	end
	if self.mode ~= mode then
		self.mode = mode
		if mode == "apply" then
			self:ApplyCurrentTemplate()
		elseif mode == "delete" then
			self:DeleteCurrentTemplate()
		end
		self:UpdateView()
	end
end

function Talented:OnInitialize()
	self.talents = {}
	self.templaterefs = {}
	
	self.mode = "view"
	
	self:InitializeOptions()
end

function Talented:RemoveTemplateFromOptions(template)
	local key = self.templaterefs[template]
	if key then
		self.options.args.template.args[key] = nil
		self.options.args.template.args.new.args[key] = nil
		self.templaterefs[template] = nil
	end
end

function Talented:DeleteCurrentTemplate()
	local template = self.template
	if template == self.current then return end
	self:RemoveTemplateFromOptions(template)
	self.db.class.templates[template.name] = nil
	self:SetTemplate(self.current)
end

function Talented:UpdateTemplateName(template, newname)
	if self.db.class.templates[newname] or self.mode == "apply" then return end
	
	local oldname = template.name
	template.name = newname
	self.db.class.templates[newname] = template
	self.db.class.templates[oldname] = nil
	local key = self.templaterefs[template]
	if key then
		self.options.args.template.args[key].name = template.name
		self.options.args.template.args[key].desc = "Select "..template.name
		self.options.args.template.args.new.args[key].name = "Copy from "..template.name
		self.options.args.template.args.new.args[key].desc = "Create a new template from "..template.name
	end
end

function Talented:AddTemplateToOptions(template)
	local count, opt, name = 1, self.options.args.template.args
	while true do
		name = string.format("template%d", count)
		if not opt[name] then break end
		count = count + 1
	end
	self.templaterefs[template] = name
	opt[name] = {
		name = template.name,
		desc = "Select "..template.name,
		type = "execute",
		order = 200,
		func = function () self:SetTemplate(template) end,
	}
	self.options.args.template.args.new.args[name] = {
		name = "Copy from "..template.name,
		desc = "Create a new template from "..template.name,
		type = "execute",
		order = 200,
		func = function() self:SetTemplate(self:CopyTemplate(template)) end,
	}
end

do
	local function new(templates, name)
		local count = 0
		local template = {
			name = name
		}
		while templates[template.name] do
			count = count + 1
			template.name = string.format("%s (%d)", name, count)
		end
		templates[template.name] = template
		return template
	end
	
	function Talented:CopyTemplate(src)
		local dst = new(self.db.class.templates, "Copy of "..src.name)
		
		for tab, talents in ipairs(self.talents) do
			local s, d = src[tab], {}
			dst[tab] = d
			for index in ipairs(talents.infos) do
				d[index] = s[index]
			end
		end
		dst.maxpoints = src.maxpoints

		self:AddTemplateToOptions(dst)
		return dst
	end

	function Talented:CreateEmptyTemplate()
		local template = new(self.db.class.templates, "Empty")
		
		for tab, talents in ipairs(self.talents) do
			local t = {}
			template[tab] = t
			for index in ipairs(talents.infos) do
				t[index] = 0
			end
		end
		template.maxpoints = self.current.maxpoints

		self:AddTemplateToOptions(template)
		return template
	end
end

function Talented:SetTemplate(template)
	self.template = template
	self.mode = "view"
	self:UpdateView()
end

function Talented:OnEnable()
	self:Hook("ToggleTalentFrame", true)
	self:SecureHook("UpdateTalentButton")
	self:RegisterEvent("CHARACTER_POINTS_CHANGED")
end

function Talented:CHARACTER_POINTS_CHANGED()
	self:UpdateCurrentTemplate()
	self:UpdateView()
	if self.mode == "apply" then
		local cp = UnitCharacterPoints("player")
		
		local reset = function(...)
			if select("#", ...) > 0 then
				self:Print(...)
			end
			self.template = self.current
			self.mode = "view"
			self:UpdateView()
			self:EnableUI(true)
		end

		local current = self.current
		local template = self.template
		for tab, talents in ipairs(self.talents) do
			local ctab = current[tab]
			local ttab = template[tab]
			for index in ipairs(talents.infos) do
				local current = ctab[index]
				local target = ttab[index]
				if current < target then
					if cp == 0 then
						reset("Error while applying talents ! Not enough character points !")
						return
					end
					self:RegisterEvent("CHARACTER_POINTS_CHANGED")
					LearnTalent(tab, index)
					return
				end
			end
		end
		reset("Template applied successfully, %d talent points remaining\n", UnitCharacterPoints("player"))
	end
end

function Talented:UpdateTalentButton()
	if self.db.profile.donthide and UnitLevel("player") < 10 then
		TalentMicroButton:Show()
		QuestLogMicroButton:SetPoint("BOTTOMLEFT", "TalentMicroButton", "BOTTOMRIGHT", -2, 0)
	end
end

function Talented:ToggleTalentFrame()
	local frame = self.base
	if not frame or not frame:IsVisible() then
		self:Update()
		ShowUIPanel(self.base)
	else
		HideUIPanel(frame)
	end
end

function Talented:Update()
	self:CreateBaseFrame()
	self:UpdateCurrentTemplate()
	if not self.template then
		self:SetTemplate(self.current)
	end
	self:UpdateView()
end
